package be.iminds.ilabt.jfed.experimenter_gui.renew_report;

import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.experimenter_gui.ui.SceneAwareController;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.util.TimeUtil;
import javafx.beans.binding.Binding;
import javafx.beans.binding.DoubleBinding;
import javafx.beans.property.ReadOnlyDoubleProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.Scene;
import javafx.scene.control.Label;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.stage.Stage;
import org.jetbrains.annotations.Contract;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.net.URL;
import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;

public class RenewReportDialog implements SceneAwareController, Initializable {
    private Slice slice;
    private ExperimentTaskStatus jobSuccessExperimentTaskStatus;
    private ExperimentController experimentController;
    private Throwable jobFailureThrowable;
    private Instant requestedRenewInstant;

    @FXML private Label summaryStatusLabel;
    @FXML private Label summaryTargetExpirationLabel;
    @FXML private Label summaryFirstExpirationLabel;
    @FXML private Label sliceExpirationLabel;
    @FXML private Label earliestSliverExpirationLabel;

    @FXML private TableColumn amColumn;
    @FXML private TableColumn sliverColumn;
    @FXML private TableColumn expirationColumn;

    @FXML private TableView<ExpirationDetail> expirationTable;
    private final ObservableList<ExpirationDetail> expirationDetails;

    private Scene scene;
    private Stage stage;

    public RenewReportDialog() {
        this.expirationDetails = FXCollections.observableArrayList();
    }

    @Override
    public void initialize(URL url, ResourceBundle resourceBundle) {
        expirationTable.setItems(expirationDetails);

        //populate table
        ReadOnlyDoubleProperty dividableWidth = expirationTable.widthProperty();
        amColumn.prefWidthProperty().bind(dividableWidth.multiply(0.25));
        sliverColumn.prefWidthProperty().bind(dividableWidth.multiply(0.375));
        expirationColumn.prefWidthProperty().bind(dividableWidth.multiply(0.375));
    }

    @Override
    public void registerScene(Scene scene) {
        this.scene = scene;
        this.stage = (Stage) scene.getWindow();
    }


    public class ExpirationDetail {
        private final String am;
        private final String sliver;
        private final String expiration;
        private final boolean ok;

        public ExpirationDetail(String am, String sliver, Instant expiration, boolean ok) {
            this(am, sliver, toHumanTime(expiration), ok);
        }
        public ExpirationDetail(String am, String sliver, String expiration, boolean ok) {
            this.am = am;
            this.sliver = sliver;
            this.expiration = expiration;
            this.ok = ok;
        }

        public String getAm() {
            return am;
        }

        public String getSliver() {
            return sliver;
        }

        public String getExpiration() {
            return expiration;
        }

        public boolean isOk() {
            return ok;
        }
    }

    public void ok() {
        if (this.stage != null)
            this.stage.close();
    }
    public void refresh() {
        //TODO: submit a job to refresh this info, and call calculate when it's done. Show a spinner while it's not.
        calculateStatus();
    }

    public void setFailStatus(String text) {
        summaryStatusLabel.setText(text);
        setLabelStyle(summaryStatusLabel, false);
    }
    public void setSuccessStatus() {
        summaryStatusLabel.setText("Successfully Renewed Slice");
        setLabelStyle(summaryStatusLabel, true);
    }

    private static final DateTimeFormatter dataTimeFormatter =
            DateTimeFormatter.ofPattern( "yyyy/MM/dd  HH:mm:ss zzz" )
                    .withZone( ZoneId.systemDefault() );
    private static String toHumanTime(@Nullable Instant instant) {
        if (instant == null)
            return "unknown";

        return dataTimeFormatter.format(instant)+"  ("+ TimeUtil.calcIn(instant)+")";
    }

    @Contract("null -> true")
    private boolean isBeforeRequestedExpireTime(@Nullable Instant actualExpirationTime) {
        if (actualExpirationTime == null)
            return true;
        if (requestedRenewInstant == null)
            throw new IllegalStateException("requestedRenewInstant is null");

        //Request times are rounded down to the nearest second. We have to take that into account.
        //We also don't care about say 5 seconds, so we'll just ignore up to a 5 second difference.
        long actualMillis = actualExpirationTime.toEpochMilli();
        long targetMillis = requestedRenewInstant.toEpochMilli();
        return (actualMillis + 5000l) < targetMillis;
    }


    private void setLabelStyle(@Nonnull Label target, boolean ok) {
        if (ok) {
            target.getStyleClass().removeAll("label-text-error", "label-error");
            target.getStyleClass().addAll("label-ok", "label-text-ok");
        } else {
            target.getStyleClass().removeAll("label-ok", "label-text-ok");
            target.getStyleClass().addAll("label-text-error", "label-error");
        }
    }

    /** Also sets style classes */
    private void setExpirationLabel(@Nonnull Label target,
                                    @Nullable Instant expireTime) {
        target.setText(toHumanTime(expireTime));
        setLabelStyle(target, !isBeforeRequestedExpireTime(expireTime));
    }

    /** this gets called after the setters, to update the status of the renew attempt */
    public void calculateStatus() {
        summaryStatusLabel.setText("?");
        summaryTargetExpirationLabel.setText("?");
        summaryFirstExpirationLabel.setText("?");
        sliceExpirationLabel.setText("?");
        earliestSliverExpirationLabel.setText("?");

        if (experimentController == null)
            throw new IllegalStateException("experimentController was not set");
        if (requestedRenewInstant == null)
            throw new IllegalStateException("requestedRenewInstant was not set");
        if (slice == null)
            throw new IllegalStateException("slice was not set");

        summaryTargetExpirationLabel.setText(toHumanTime(requestedRenewInstant));
        setExpirationLabel(sliceExpirationLabel, slice.getExpirationDate());

        boolean allOk = !isBeforeRequestedExpireTime(slice.getExpirationDate());

        Instant earliestSliverExpiration = null;
        List<ExpirationDetail> newExpirationDetails = new ArrayList<>();
        for (Sliver sliver: slice.getSliversCopy()) {
            newExpirationDetails.add(new ExpirationDetail(
                    sliver.getServer().getName(),
                    sliver.getUrnString(),
                    sliver.getExpirationDate(),
                    !isBeforeRequestedExpireTime(sliver.getExpirationDate())
            ));
            if (sliver.getExpirationDate() != null) {
                if (earliestSliverExpiration == null ||
                        earliestSliverExpiration.isAfter(sliver.getExpirationDate()))
                    earliestSliverExpiration = sliver.getExpirationDate();
                allOk = allOk && !isBeforeRequestedExpireTime(slice.getExpirationDate());
            } else {
                allOk = false;
            }
        }
        setExpirationLabel(earliestSliverExpirationLabel, earliestSliverExpiration);
        Instant earliestSliverOrSliceExpiration =
                earliestSliverExpiration == null ?
                        slice.getExpirationDate() :
                        (slice.getExpirationDate() == null ? earliestSliverExpiration :
                                (slice.getExpirationDate().isBefore(earliestSliverExpiration) ?
                                        slice.getExpirationDate() : earliestSliverExpiration)
                        );
        setExpirationLabel(summaryFirstExpirationLabel, earliestSliverOrSliceExpiration);
        allOk = allOk && !isBeforeRequestedExpireTime(earliestSliverExpiration);

        expirationDetails.setAll(newExpirationDetails);

        //noinspection VariableNotUsedInsideIf
        if (jobFailureThrowable != null) {
            allOk = false;
            setFailStatus("Job Failed - Status Unknown");
            return;
        }
        if (jobSuccessExperimentTaskStatus == null) {
            allOk = false;
            setFailStatus("Job Status Unknown");
            return;
        }
        if (jobSuccessExperimentTaskStatus != ExperimentTaskStatus.SUCCESS) {
            allOk = false;
            setFailStatus("Job was not successfull: "+jobSuccessExperimentTaskStatus);
            return;
        }

        if (allOk) {
            setSuccessStatus();
        } else {
            setFailStatus("Renew failed.");
        }
    }

    public void setJobSuccessExperimentTaskStatus(@Nullable ExperimentTaskStatus jobSuccessExperimentTaskStatus) {
        this.jobSuccessExperimentTaskStatus = jobSuccessExperimentTaskStatus;
    }

    public void setExperimentController(@Nonnull ExperimentController experimentController) {
        this.experimentController = experimentController;
        this.slice = experimentController.getExperiment().getSlice();
    }

    public void setJobFailureThrowable(@Nonnull Throwable jobFailureThrowable) {
        this.jobFailureThrowable = jobFailureThrowable;
    }

    public void setRequestedRenewInstant(@Nonnull Instant requestedRenewInstant) {
        this.requestedRenewInstant = requestedRenewInstant;
    }
}
